import os
import json
from slogger import Logger
import demjson
from fn import F
from .defaults import write as write_defaults, all as all_defaults
from .roles import build as roles_build, link as roles_link, load as roles_load, link_src_to_deploy
from .tools.file import put_file
from .tools.folder import put_folder
from .tools.dicthelper import update_dict
from .util import get_config_path, get_deploy_path, get_deploy_host, get_deploy_ansible_cfg_path
from .deploy_info import DeployInfo

logger = Logger.getLogger(__name__)

def run_deploy(project_name, config_name, tags=None):
    '''configuration_path: it is the folder path which contains the main.sh to launch the deploy script'''
    import subprocess

    def run(deployInfo):
        def build_base_command():
            return "sudo ansible-playbook ./{yml_file} -i ./{host_file}".format(yml_file=deployInfo.playbook_name(), host_file=deployInfo.host_file_name())

        def build_tag(base_command):
            return base_command if not tags else \
                '{base_command} --tags "{tags}"'.format(base_command=base_command, tags=",".join(tags))
        
        p = subprocess.Popen(logger.title('cli').debug((F(build_base_command) >> F(build_tag))()), shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=put_folder(deployInfo.deploy_folder_path()))

        while p.poll() is None:
            line = p.stdout.readline()
            if line:
                print(line.strip())

    run(DeployInfo(get_deploy_path(project_name, config_name)))

def build_deploy_script(project_name, config_name, only_structure=False, remote_dict=None, gate=False, proxy_mapping=None, noauth=None, authorization=None):
    from .gate import build_gate
    
    build_gate(project_name, config_name, gate, proxy_mapping, noauth, authorization)
    build_deploy_script_internal(project_name, config_name, only_structure, remote_dict)
    
def build_deploy_script_internal(project_name, config_name, only_structure=False, remote_dict=None):
    deploy_path = lambda:get_deploy_path(project_name, config_name)
    deployInfo = DeployInfo(deploy_path())
    
    def yml_file_folder():
        return put_folder(os.path.abspath(deploy_path()))
    def get_roles_data():
        return roles_load(logger.title("config_path").debug(get_config_path(project_name, config_name)))
    def write_playbook(content):
        open(put_file(deployInfo.playbook_path()), 'w').write(content)

    # write data to file
    write_playbook( \
                    roles_build( \
                                 get_roles_data() \
                                 , remote_host=None if remote_dict is None else 'remote' \
                                 , remote_name=remote_dict["remote_user"] if remote_dict else None \
                    ) \
    )
    # build link for role folders

    roles_link(get_roles_data(), \
               logger.title("link_root").debug(yml_file_folder()) \
    )

    # link src folder to deploy/roles/main/files/src for deployment
    link_src_to_deploy(get_roles_data())

    def default_vals():
        return all_defaults(yml_file_folder(), [role["name"] for role in get_roles_data()])

    # build inventory files on roles
    write_defaults(deployInfo.host_file_path(), default_vals(), remote_addr=None if remote_dict is None else remote_dict["remote_addr"])

    # write the defaults to cfg file
    (F(lambda config_path: demjson.decode_file(config_path)) >> \
        F(lambda new_vals, old_json: update_dict(old_json, new_vals), {} if only_structure else {"predefined_variables": default_vals()}) >> \
        F(lambda json_data: json.dumps(json_data, indent=4, ensure_ascii=False)) >> \
        F(lambda content: open(get_config_path(project_name, config_name), 'w').write(content)))(get_config_path(project_name, config_name))


    def build_ansible_cfg():
        if remote_dict is None:
            return

        open(get_deploy_ansible_cfg_path(project_name, config_name), 'w').write('''[defaults]
remote_user={remote_user}
private_key_file={key_file}
host_key_checking=False
'''.format(remote_user=remote_dict["remote_user"], key_file=remote_dict["key"]))
        
    build_ansible_cfg()


def deploy_config(project_name, config_name):
    import configparser
    config = configparser.ConfigParser(allow_no_value=True)
    config.read(get_deploy_host(project_name, config_name))
    
    def get_banyan_config():
        return demjson.decode_file(get_config_path(project_name, config_name))
    def get_predefined_variables(banyan_config):
        return banyan_config["predefined_variables"] if "predefined_variables" in banyan_config else {}

    def set_config(updatedDict):
        config["all:vars"] = updatedDict
        return config
    
    def save_items_to_host(cfg):
        cfg.write(open(get_deploy_host(project_name, config_name), 'w'))

    (F(get_banyan_config) >> \
        F(get_predefined_variables) >> \
        F(update_dict, dict(config["all:vars"])) >> \
        F(set_config) >> \
        F(save_items_to_host))()

    





import os
import shutil
from .assert_util import not_none as assert_not_none

env_dict = {}

def init_root(path = None):
    env_dict["root_path"] = os.getcwd() if path == None else os.path.abspath(path)
    
def get_root():
    return env_dict["root_path"]

def extract_project_path(config_path):
    import re
    def match(pattern):
        return re.match(pattern, config_path)[0] if re.match(pattern, config_path) else None
    return os.path.abspath(assert_not_none(match('''.*(?=/deploy/\w+\.cfg)'''), "cfg path is invalid. Please make sure the it is under project folder"))

def get_project_path(project_name):
    return os.path.abspath(os.path.join(get_root(), project_name))

def get_config_path(project_name, config_name = None):
    return os.path.abspath(os.path.join(get_root(), project_name, "deploy", "dev.cfg" if config_name == None else config_name + ".cfg"))

def get_deploy_path(project_name, config_name = None):
    return os.path.abspath(os.path.join(get_root(), project_name, "deploy", ".dev" if config_name == None else "." + config_name))

def get_deploy_host(project_name, config_name = None):
    return os.path.join(get_deploy_path(project_name, config_name), "main.host")

def get_deploy_yml(project_name, config_name = None):
    return os.path.join(get_deploy_path(project_name, config_name), "main.yml")

def get_deploy_ansible_cfg_path(project_name, config_name = None):
    return os.path.join(get_deploy_path(project_name, config_name), "ansible.cfg")



from codegenhelper import remove
import os

def clear_deploy(projectpath, config_name):
    def clear(path):
        if os.path.exists(path):
            remove(path)
    clear(os.path.join(projectpath, "deploy", "." + config_name))



import os
import demjson
from .assert_util import not_none as assert_not_none
from slogger import Logger
from .util import get_project_path, extract_project_path, get_config_path
from pathlib import Path
from fn import F
from .config import load_all_config

logger = Logger.getLogger(__name__)


def load(config_path):
    def convert(data):
        return {
            "name": "{}_{}".format(data["project_name"], data["role_name"]),
            "local": data["project_path"],
            "original_name": data["role_name"]
        }

    return [convert(data) for data in load_all_config(config_path)]

def build(roles, remote_host = None, remote_name = None):
    from jinja2 import Template
    def get_host(): return 'hosts: localhost' if remote_host == None else 'hosts: ' + remote_host
    def get_name(): return '''environment:
    PYTHONPATH: "/home/{remote_name}/.local/lib/python2.7/site-packages" '''.format(remote_name = remote_name) if remote_name else ""
    '''
build the roles into the entry yaml file
roles: a list for role. the sturcture of role can be reference in get_roles
'''
    return Template('''
---

- name: deploy
  {{ hosts }}
  become: true
  become_method: sudo
  {{ remote_name }}
  roles:
{% for role in objs %}    - {'role':'{{role}}', 'tags':'{{role}}'}
{% endfor %}
...
  
''').render(objs = [role["name"] for role in roles], hosts = get_host(), remote_name = get_name())


def link(roles, path):
    '''
link the roles
'''
    def get_src(role):
        return os.path.join(role["local"], "deploy", "roles", role["original_name"])

    def create_folder(name, parent_path):
        def create(path):
            if os.path.exists(path):
                return path
            os.makedirs(path)
            return path
        return create(os.path.join(parent_path, name))
    
    def get_dest(role):
        return os.path.join(create_folder("roles", os.path.abspath(path)) , role["name"])
    
    def link_role(role):
        if not os.path.exists(get_dest(role)):
            os.symlink(os.path.abspath(get_src(role)), get_dest(role))
        
    [link_role(role) for role in roles]


def link_src_to_deploy(rolesdata):
    ''' link src folder to deploy/roles/main/files/src for deployment '''
    
    def link(project_folder, src_name, link_folder):
        def put(target_path):
            if not os.path.exists(target_path):
                os.makedirs(target_path)
            return target_path
        
        def link_path(source_path, target_link_path):
            if not os.path.exists(source_path): return

            if Path(target_link_path).is_symlink(): os.remove(target_link_path)

            if os.path.exists(target_link_path): raise ValueError("the path %s can't be removed"% target_link_path)

            os.symlink(source_path, target_link_path)

        link_path(os.path.abspath(os.path.join(project_folder, src_name)),  os.path.join(\
                                                                                         put(os.path.abspath(os.path.join(project_folder, link_folder))),\
                                                                                         src_name)\
        )
        
    [link(project_folder, "src", "deploy/roles/main/files") for project_folder in list(set([p["local"] for p in rolesdata]))]
        



import os
import re

def encode_val(val):
    return val if type(val) is not str else '''"%s"''' % val
    
def dict2yml(dict_data):
    
    def gen_item(key, val):
        return "%s: %s" % (key, encode_val(val))
    
    return os.linesep.join([gen_item(key, val) for key, val in dict_data.items()]) if isinstance(dict_data, dict) else ""

def dict2assignments(dict_data):
    return ['''%s=%s''' % (key, val) \
                for key, val in dict_data.items()]



from .config import load_all_config
from .util import get_project_path, extract_project_path, get_config_path
import os
import sys
from slogger import Logger
import demjson
from fn import F
import json
from functools import reduce
from assertpy import assert_that

logger = Logger.getLogger(__name__)

_auth_db_name = "auth_db"
_microservice_gate_name = "microservice_gate"

def build_gate(project_name, config_name, build_gate, proxy_mapping = None, noauth = None, authorization = None):
    def remove_duplicate(projects):
        def work(result, item):
            if isinstance(result, list):
                if len(list(filter(lambda obj:obj['project_name'] == item['project_name'], result))) != 0:
                    return result
                else:
                    return result + [item]
            else:
                if result['project_name'] == item['project_name']:
                    return [result]
                else:
                    return [result, item]
                
                
        return projects if len(projects) == 1 else [] if len(projects) == 0 else reduce(work, projects)
    

    def build_authorization(projects):
        '''
            authorization is one of the service
        '''
        def get_authorized_project():
            return list(filter(lambda n: n['project_name'] == authorization, projects))

        def get_proxy_mapping_projects():
            return list(filter(lambda n:n['has_customized_mapping'], projects))

        def get_no_auth_projects():
            return list(filter(lambda n:n['noauth'], projects))

        try:

            if authorization:
                return get_authorized_project() + \
                               get_proxy_mapping_projects() + \
                               get_no_auth_projects()
            else:
                return projects
        except Exception as ex:
            logger.title('build_authorization.error').error(ex)
            logger.title('build_authorization.error.data').error(projects)
            
            raise ex

    
    def data():
        def dumy(projects):
            assert_that(projects).is_not_empty()
            assert_that(projects[0]).contains_key('project_name', 'project_path', 'role_name')
            
            return {
                "project_name": "roles",
                "items": (F(remove_duplicate) >> \
                          F(proxy_mapping) >> \
                          F(root_path_at_last) >> \
                          F(no_auth) >> \
                          F(build_authorization) >> \
                          F(remove_duplicate) \
                )(projects)
            }
        return dumy(load_all_config(get_config_path(project_name, config_name)))

    def build_no_auth_list():
        return [] if noauth == None else \
            noauth if isinstance(noauth, list) else \
            [noauth]
    
    def no_auth(all_config, no_auth_list = build_no_auth_list()):
        def build(item):
            item['noauth'] = True if item['project_name'] in no_auth_list else False
            return item

        return list(map(lambda item: build(item), all_config))
    def build_mapping():
        '''convert the data
[a:b] to 
[[a,b]] '''
        return [] if proxy_mapping == None else \
            list(map(lambda item: [item.split(':')[0], item.split(':')[1]], proxy_mapping if isinstance(proxy_mapping, list) else [proxy_mapping]))
    
    def proxy_mapping(all_config, mapping = build_mapping()):
        def get_mapping(item):
            return list(filter(lambda obj: obj[0] == item['project_name'], mapping))
        
        def build(item):
            return {
                'project_name': item['project_name'],
                'project_path': item['project_path'],
                'role_name': item['role_name'],
                'proxy_mapping': '/_api/{}/'.format(item['project_name']) if len(get_mapping(item)) == 0 else get_mapping(item)[0][1],
                'has_customized_mapping': len(get_mapping(item)) > 0
            }

        return list(map(lambda item: build(item), all_config))

    def root_path_at_last(all_config):
        '''make sure the root path / is at the last'''
        sorted_list = sorted(all_config, key = lambda n:n['proxy_mapping'])
        sorted_list.reverse()
        return sorted_list
        
    def output_path():
        return os.path.join(get_project_path(project_name), "deploy")

    def template_path():
        return os.path.join(os.path.split(os.path.realpath(__file__))[0], "./auth_template")

    def open_config():
        return demjson.decode_file(get_config_path(project_name, config_name))

    def update_config(remove, config_data):
        roles_seq = [item for item in logger.title("origin_roles_seq").debug(config_data["roles_seq"]) if item not in [_auth_db_name, _microservice_gate_name]]
        
        config_data["roles_seq"] = logger.title("pure_roles_seq").debug(roles_seq) + ([] if remove else [_auth_db_name, _microservice_gate_name])
        
        return config_data

    def save_config(config_data):
        try:
            with open(get_config_path(project_name, config_name), 'w') as f:
                json.dump(config_data, f, indent=4)
        except Exception as e:
            logger.title('save_config.error').error(e)
            logger.title('save_config.data').error(config_data)
            
    def modify_config(remove):
        (F(open_config) >>\
            F(update_config, remove) >>\
            F(save_config)
         )()
    def gen_template():
        from md_codegen import run
        run(logger.title('gen_template.output_path').debug(output_path()), \
            "",\
            None,\
            None,\
            None, \
            None,\
            None,\
            demjson.encode(data()),\
            None,\
            template_path(),\
            check_repo = False)

        modify_config(False)
        
    def remove_gate():
        import shutil

        def remove_folder(roles_folder):
            [lambda path: shutil.rmtree(os.path.join(roles_folder, path)) for path in [_auth_db_name, _microservice_gate_name] \
             if os.path.exists(os.path.join(roles_folder, path))]
            
        remove_folder(os.path.join(get_project_path(project_name), "deploy", "roles"))
        
        modify_config(True)

    try:
        
        if logger.title('build_gate').debug(build_gate) == True:
            remove_gate()
            gen_template()
        else:
            remove_gate()
    except Exception as e:
        import traceback
        logger.title('build_gate.error.trackback').error(''.join(traceback.format_tb(e.__traceback__)))
        logger.title('build_gate.error').error(e)

        




import os
from slogger import Logger
import subprocess
import re
from assertpy import assert_that

logger = Logger.getLogger(__name__)

def get_project_name_from_url(giturl):
    def parseName(basename):
        return basename.split(".")[0] if len(basename.split(".")) == 2 and basename.split(".")[1] == "git" else basename
    return parseName(os.path.basename(giturl))

def get_tag(giturl, tag, location = os.getcwd(), ssh_key = None):
    assert_that(location).exists()
    assert_that(tag).is_not_empty()
    assert_that(giturl).is_not_empty()
    
    def tag_cmd():
        return '-b "%s"' % tag

    def key_cmd():
        return "" if ssh_key == None else \
            'GIT_SSH_COMMAND="ssh -i %s"'%os.path.abspath(ssh_key)

    def git_cmd():
        return "git clone --single-branch --depth 1 %s" % giturl

    subprocess.call(logger.title('get_tag cmd').debug(" ".join([key_cmd(), git_cmd(), tag_cmd()])), shell=True, cwd=logger.title('current path').debug(location))
    
def get_code(giturl, ssh_key = None, location = os.getcwd()):
    assert_that(giturl).is_not_empty()
    assert_that(location).exists()
    def is_ssh():
        return giturl.startswith('git')
    
    subprocess.call(logger.title('get_code ssh cmd').debug('ssh -i {key_file} git clone {giturl}'.format(key_file=ssh_key, giturl=giturl)), \
                    shell=True, \
                    cwd=location) if is_ssh() else \
                    subprocess.call(logger.title('get_code cmd').debug('git clone {giturl}'.format(giturl=giturl)), \
                                    shell=True, \
                                    cwd=location)
    
    



#!/usr/bin/env python
from setuptools import setup, find_packages
name = "mc_banyan"

requires = ['demjson>=2.2.4', 'jinja2>=2.10', 'pyyaml>=3.13', 'shell>=1.0.1', 'codegenhelper>=0.0.13', 'fn', 'md_codegen>=0.5.0', 'slogger>=1.2.1', 'requests>=2.21.0', 'assertpy>=0.14']

setup(
    name = name,
    version = '2.3.1',
    scripts = ["scripts/banyan"],
    author = 'Zongying Cao',
    author_email = 'zongying.cao@dxc.com',
    description = 'banyan is a tool to manage the deployment tasks for microservices.',
    long_description = """banyan is a tool to manage the deployment tasks for microservices.""",
    packages = [name],
    package_dir = {'mc_banyan': 'src'},
    package_data = {'mc_banyan': ["*.py", "*.t"]},
    include_package_data = True,
    install_requires = requires,
    license = 'Apache',
    classifiers = [
               'Development Status :: 4 - Beta',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: Apache Software License',
               'Natural Language :: English',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 3',
               'Topic :: Software Development :: Libraries',
           ],
)



import os

class DeployInfo:
    def __init__(self, deploy_folder_path):
        self.playbook_name = lambda: 'main.yml'
        self.host_file_name = lambda: 'main.host'

        self.playbook_path = lambda: os.path.join(deploy_folder_path, self.playbook_name())
        self.host_file_path = lambda: os.path.join(deploy_folder_path, self.host_file_name())

        self.deploy_folder_path = lambda: deploy_folder_path



import os
from fn import F
import demjson
from slogger import Logger
from .util import get_project_path, extract_project_path, get_config_path

logger = Logger.getLogger(__name__)

def load_all_config(config_path):
    '''
config_path: the path of the banyan configuration file
'''
    def get_cfg(dep_json):
        def load_cfg(cfg_path):
            return demjson.decode_file(cfg_path)

        return (F(get_config_path) >> F(load_cfg))(dep_json["project_name"], dep_json["config_name"] if "config_name" in dep_json else None)

    def get_roles(cfg_json, project_path):
        return [{"project_name": os.path.split(project_path)[1], \
                 "project_path": project_path,
                 "role_name": role_name} for role_name in cfg_json["roles_seq"]] if "roles_seq" in logger.title("get_roles cf_json").debug(cfg_json) else []
    
    def handle(cfg_json, depended_roles, project_path):
        try:
            return [y for x in [handle(get_cfg(dep), depended_roles, get_project_path(dep["project_name"])) for dep in cfg_json["dependencies"]] for y in x] \
                + get_roles(cfg_json, logger.title("get_roles_path_recursive").debug(project_path)) \
            if "dependencies" in cfg_json \
               else depended_roles + get_roles( \
                                                logger.title("no dep json").debug(cfg_json) \
                                                , logger.title("no dep path").debug(project_path) \
               )
        except Exception:
            logger.error("error happen in handle with project_path:%s\ncfg_json:%s" % (project_path, cfg_json))
            raise

    def put_auth_db_role_first(roles):
        def change_position(auth_db_roles):
            assert not auth_db_roles or len(auth_db_roles) == 1, "auth_db is not correctly setup"
            
            return auth_db_roles \
                + roles[0: roles.index(auth_db_roles[0])] \
                + roles[roles.index(auth_db_roles[0]) + 1:] \
                if auth_db_roles else roles
        
        return change_position( \
                                list( \
                                      filter( \
                                              lambda n: n['role_name'] == 'auth_db' \
                                              , roles \
                                      ) \
                                ) \
        )
    
    return put_auth_db_role_first( \
                                   handle( \
                                           demjson.decode_file(config_path) \
                                           , [] \
                                           , extract_project_path(config_path) \
                                   ) \
                                   )
                                   

def load_all_dependencies(config_path, handler):
    dependencies = lambda: "dependencies"
    
    def handle(dep_data):
        def handle_config(config_data):
            if dependencies() in config_data:
                [handle(dep) for dep in config_data[dependencies()]]
                
        handle_config(handler(dep_data))

    def handle_from_root(config_data):
        if dependencies() in config_data:
            [handle(dep) for dep in config_data["dependencies"]]

    handle_from_root(demjson.decode_file(config_path))
